import pandas as pd
import numpy as np
import h5py
from scipy.sparse import csr_matrix
from sknetwork.classification import DiffusionClassifier
from sknetwork.regression import Dirichlet
from projectframe import ProjectFrame
from joblib import Parallel, delayed
import os

# Local imports
from ...data.import_data import get_data_file
from ...core.utilities import to_microarray


class EPICimputer(object):
    """
    """
    
    def __init__(self, network_file="EPIC_probe_network.h5"):
        """
        """

        # Get File
        #if not os.path.exists(network_file):
        #    network_file = get_data_file(network_file)
        network_file = "/opt/data/EPIC_probe_network.h5"
        
        # Read imputation matrix
        f = h5py.File(network_file, "r")
        self.features = np.array(f["names"]).astype(str)
        data = np.array(f["data"])
        indices = np.array(f["indices"])
        indptr = np.array(f["indptr"])
        shape = np.array(f["shape"])
        shape = (shape[0], shape[1])
        f.close()
        self.network = csr_matrix((data, indices, indptr), shape=shape)
        
        
    def match_matrix(self, X: pd.DataFrame):
        """
        """

        # Check if the features match
        if np.array_equal(self.features, X.columns.values):
            return X
        
        # Copy
        X_copy = X.copy()
        
        # Match features
        common = X_copy.columns.intersection(self.features)
        if len(common) == 0:
            common = X_copy.index.intersection(self.features)
            if len(common) == 0:
                raise ValueError("No matching features found.")
            X_copy = X_copy.T

        missing = pd.Index(self.features).difference(X_copy.columns.values)
        X_copy = X_copy.loc[:,common]
        #X_copy.loc[:,missing] = np.nan
        #for feature in missing:
        #    X_copy.loc[:,feature] = np.nan
        null_df = pd.DataFrame(np.nan, index=X_copy.index, columns=missing)
        X_copy = pd.concat([X_copy, null_df], axis=1)
        X_copy = X_copy.loc[:,self.features]
        
        return X_copy
        
    
    def impute_status(self, X, threshold=0.75):
        """
        """

        # Initialize the prediction matrix
        prediction = self.match_matrix(X)
    
        # Fit the Diffusion model
        diffusion = DiffusionClassifier()
    
        # Iterate over the columns
        for n in range(prediction.shape[0]):            
            #print(n, flush=True)

            # Predict the missing values
            x = prediction.values[n,:]
            is_null = pd.isnull(x)
            if np.sum(~is_null) > 10:
                labels = {i:x[i] for i in range(len(x)) if np.isnan(x[i]) == False}
                labels_pred = diffusion.fit_predict(self.network, labels)
                probs = diffusion.predict_proba().max(axis=1)

                confident = np.logical_and(is_null, probs >= threshold)
                prediction.iloc[n,confident] = labels_pred[confident]

        return prediction
        
    
    def impute_beta(self,
                    X: pd.core.frame.DataFrame,
                    verbose: bool = False):
        """
        Predict the missing values in X using the network A.

        Parameters
        ----------
            X : pd.DataFrame
                The data matrix.
            verbose : bool
                If True, print progress.

        Returns
        -------
            prediction : pd.DataFrame
                The predicted values.
        """

        # Initialize the prediction matrix
        prediction = self.match_matrix(X)

        # Fit the Dirichlet model
        dirichlet = Dirichlet()

        # Iterate over the columns
        output_values = prediction.values
        for n in range(prediction.shape[0]):
            if verbose: print(n, flush=True)

            # Predict the missing values
            x = prediction.values[n,:]
            if np.sum(~pd.isnull(x)) > 10:
                #values = {i:x[i] for i in range(len(x)) if np.isnan(x[i]) == False}
                values = {i: x[i] for i in np.where(~np.isnan(x))[0]}
                values_pred = dirichlet.fit_predict(self.network, values)
            else:
                values_pred = np.repeat(np.nan, len(x))

            output_values[n,:] = values_pred
        
        prediction.loc[:,:] = output_values

        return prediction


def chunkify(data, chunks):
    return np.array_split(data, chunks)

def parallel_impute_status(imputer, data, threshold, n_jobs=-1, chunks=10):
    data_chunks = chunkify(data, chunks)
    results = Parallel(n_jobs=n_jobs)(delayed(imputer.impute_status)(chunk, threshold) for chunk in data_chunks)
    return pd.concat(results, axis=0)

def parallel_impute_beta(imputer, data, n_jobs=-1, verbose=False, chunks=10):
    data_chunks = chunkify(data, chunks)
    results = Parallel(n_jobs=n_jobs)(delayed(imputer.impute_beta)(chunk, verbose) for chunk in data_chunks)
    return pd.concat(results, axis=0)

def impute_epic(data: pd.DataFrame,
                network_file: str = "EPIC_probe_network.h5",
                method: str = "binary",
                threshold: float = 0.75,
                n_imputes: int = 5,
                fill: bool = False,
                n_jobs: int = -1,
                chunks: int = 10):
    """
    Perform imputation using multiprocessing in chunks to optimize memory usage.
    
    Parameters
    ----------
        pf : ProjectFrame
            ProjectFrame object containing data.
        network_file : str
            Path to the network file.
        method : str
            "binary" for binary classification or "continuous" for Dirichlet regression.
        threshold : float
            Confidence threshold for binary imputation.
        n_imputes : int
            Number of iterative imputations.
        fill : bool
            Whether to fill remaining missing values with threshold=0.0.
        n_jobs : int
            Number of parallel jobs (-1 uses all available cores).
        chunks : int
            Number of chunks to split data for parallel processing.
    
    Returns
    -------
        pf : ProjectFrame
            Updated ProjectFrame with imputed values.
    """
    
    # Read network
    imputer = EPICimputer(network_file=network_file)
    
    if method == "binary":
        data[data > 0.6] = 1
        data[data <= 0.6] = 0

        # Impute in parallel using chunks
        for _ in range(n_imputes):
            data = parallel_impute_status(imputer, data, threshold, n_jobs, chunks)
        
        if fill:
            data = parallel_impute_status(imputer, data, threshold=0.0, n_jobs=n_jobs, chunks=chunks)
        
        # Format
        data[data == 0] = -1
        data[pd.isnull(data)] = 0
    
    else:
        # Continuous imputation in parallel using chunks
        data = parallel_impute_beta(imputer, data, n_jobs, chunks=chunks)
    
    return data


#from ailist import LabeledIntervalArray
#from cfdna import cfDNA
from statistics import median
from ..data.import_data import get_data_file
from intervalframe.read.read_h5 import read_h5_intervalframe
from intervalframe.read.read_text import read_bed
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray
from projectframe import ProjectFrame
import h5py
import pandas as pd
import numpy as np
import glob
import os


def predict_data_type(directory: str) -> str:
    """
    Predict data type from directory

    Parameters
    ----------
        directory : str
            Directory containing data

    Returns
    -------
        data_type : str
            Data type
    """

    # Check if bedgraph files in directory
    if os.path.isdir(directory):
        bed_files = glob.glob(directory + "**/*.bedGraph", recursive=True)
        if len(bed_files) > 0:
            return "sequencing_directory"
        else:
            return "microarray_directory"
    
    # Check if directory is idat file
    if os.path.isfile(directory + "_Grn.idat"):
        return "microarray_single"
    
    # Check if directory is bed file
    if os.path.isfile(directory + ".bedGraph") or directory.endswith(".bedGraph"):
        return "sequencing_single"

    return "unknown"    


def position_to_cpg(iframe: IntervalFrame,
                    genome_version: str = "hg38"):
    
    # Read annotation file
    if genome_version == "hg38":
        #anno_file = get_data_file("hg38_anno.parquet")
        anno_file = "./classifier_data/methyl_data_v2/microarray_data/hg38_anno.parquet"
    else:
        #anno_file = get_data_file("hg19_anno.parquet")
        anno_file = "./classifier_data/methyl_data_v2/microarray_data/hg19_anno.parquet"
    iframe_cn = IntervalFrame.read_parquet(anno_file)
    probes = iframe_cn.df.loc[:,"cpg_name"].values

    # Overlap
    cpg_name = np.empty(iframe.shape[0], dtype="U25")
    for i, index in enumerate(iframe_cn.index.iter_intersect(iframe.index, return_intervals=False, return_index=True)):
        if len(index > 0):
            cpg_name[i] = probes[index[0]]

    return cpg_name


def get_probe_names():
    # Read annotation file
    #iframe_cn = IntervalFrame.read_parquet(get_data_file("hg19_anno.parquet"))
    iframe_cn = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg19_anno.parquet")
    probes = iframe_cn.df.loc[:,"cpg_name"].values

    return probes


def convert_v2_to_v1(cpgs: pd.DataFrame) -> pd.DataFrame:
    """
    """

    # Read conversion file
    v2_to_v1 = pd.read_parquet(get_data_file("EPIC2_to_EPIC.parquet"))
    v2_to_v1 = pd.read_parquet("./classifier_data/methyl_data_v2/microarray_data/EPIC2_to_EPIC.parquet")

    # Convert
    #new_probes = cpgs.columns.intersection(v2_to_v1.index.values)
    #cpgs = cpgs.loc[:,common_probes]
    #cpgs.columns = v2_to_v1.loc[common_probes,"EPICv1_Loci"].values
    cpgs.rename(columns=v2_to_v1.to_dict()["EPICv1_Loci"], inplace=True)

    return cpgs


def cpg_to_position(cpgs,
                    genome_version: str = "hg38"):
    """
    """

    # Read annotation file
    if genome_version == "hg38":
        iframe_cn = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg38_anno.parquet")
        iframe_cn = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg38_anno.parquet")
    else:
        iframe_cn = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg19_anno.parquet")
    probes = iframe_cn.df.loc[:,"cpg_name"].values

    # Build cn probes
    cn_probes = pd.Series(np.arange(iframe_cn.shape[0]),
                          index=iframe_cn.df.loc[:,"cpg_name"].values)

    # Find common
    common_probes = pd.Index(cpgs).intersection(probes)
    iframe_cn = iframe_cn.iloc[cn_probes.loc[common_probes].values,:]

    return iframe_cn


def bin_cpgs(cpg_iframe,
             bins = None,
             bin_size=100000,
             method="mean"):
    """
    """

    # Chosen function
    if method == "mean":
        func = np.mean
    elif method == "sum":
        func = np.sum
    elif method == "median":
        func = np.median
    elif method == "std":
        func = np.std
    elif method == "nanmean":
        func = np.nanmean
    elif method == "nansum":
        func = np.nansum
    elif method == "nanmedian":
        func = np.nanmedian

    if bins is None:
        # Get ranges
        ranges = cpg_iframe.index.label_ranges
        range_dict = {chrom:ranges[chrom][1] for chrom in ranges}

        # Create bins
        bin_iframe = IntervalFrame.from_dict_range(range_dict, bin_size)
    else:
        bin_iframe = bins

    values = pd.DataFrame(np.zeros((bin_iframe.df.shape[0],cpg_iframe.shape[1])),
                          columns = cpg_iframe.columns.values)
    values[:] = np.nan
    for i, index in enumerate(cpg_iframe.index.iter_intersect(bin_iframe.index, return_intervals=False, return_index=True)):
        value = cpg_iframe.df.iloc[index,:].values
        values.values[i,:] = func(value, axis=0)

    bin_iframe.df = values
    
    return bin_iframe


def mean_methyl(intervals, cpg_iframe):
    """
    """

    bin_iframe = IntervalFrame(intervals=intervals.index.copy())

    values = pd.DataFrame(np.zeros((bin_iframe.df.shape[0],cpg_iframe.shape[1])),
                          columns = cpg_iframe.columns.values)
    for i, index in enumerate(cpg_iframe.index.iter_intersect(bin_iframe.index, return_intervals=False, return_index=True)):
        value = cpg_iframe.df.values[index,:]
        values.values[i,:] = np.mean(value, axis=0)

    bin_iframe.df = values
    
    return bin_iframe


def methyl_windows(intervals, cpg_iframe, obs):
    """
    """

    size = intervals.index[0].end - intervals.index[0].start
    obs_cpgs = cpg_iframe.loc[:,obs]
    
    total_values = np.zeros((intervals.shape[0], size))
    for i, interval in enumerate(intervals.index):
        cpgs = obs_cpgs.intersect(interval.start, interval.end, interval.label)
        values = np.zeros(interval.end - interval.start)
        values[:] = np.nan
        values[cpgs.index.starts - interval.start] = cpgs.values

        total_values[i,:] = values

    return total_values


def mean_bigwig(intervals, bigwig_fn):
    """
    """
    import pyBigWig
    bw = pyBigWig.open(bigwig_fn)

    means = np.zeros(len(intervals))
    for i, interval in enumerate(intervals):
        values = bw.values(interval.label, interval.start, interval.end, numpy=True)
        means[i] = np.nanmean(values)

    iframe = IntervalFrame(intervals=intervals, df=pd.DataFrame(means))
    iframe.df.columns = ["mean"]

    return iframe


def to_microarray(pf: ProjectFrame,
                    genome_version: str = "hg38") -> ProjectFrame:
    """
    """
    betas = pf.intervals["betas"]
    cpgs = position_to_cpg(betas, genome_version=genome_version)
    present = cpgs != ""
    filtered_betas = betas.df.loc[present,:]
    filtered_betas.index = cpgs[present]
    filtered_betas = filtered_betas.T
    filtered_betas = filtered_betas.loc[:,~filtered_betas.columns.duplicated()]

    pf.add_values("microarray_betas", filtered_betas)

    return pf


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.gridspec as gridspec
import seaborn as sns
from scipy.stats import pearsonr, spearmanr, sem
from scipy.signal import argrelmax
from ngsfragments.correct.correction import gaussian_smooth
import os
from projectframe import ProjectFrame

# Tell matplot lib to use 'editable text'
import matplotlib
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype'] = 42


def plot_table(sam_fn, genome, bin_size, purity, ploidy, clonal, n_frags, nfr,
               median_var, font_size=14, show=True, save=None, ax=None):
    """
    Plot table of summary metrics
    """

    # Create Axes object if not given
    if ax == None:
        with sns.axes_style("ticks"):
            fig, ax = plt.subplots(figsize=(7,5), tight_layout=True)

    # Record text
    labels = ["SAM file name:", "Tumour purity:",
              "Ploidy:", "Clonality:",
              "Number of fragments:",
              "NFR:", "Median variance:",
              "Genome:", "Bin size:"]
    numbers = [sam_fn, '{:.3f}'.format(purity),
               '{:.3f}'.format(ploidy), '{:.3f}'.format(clonal),
               str(n_frags), '{:.3f}'.format(nfr),
               '{:.3f}'.format(median_var),
               genome, str(bin_size)]
    text = list(zip(labels, numbers))

    # Table formatting
    color = "lightgrey"
    blank = "white"
    cellColours = [[color, color],
                   [blank, blank],
                   [color, color],
                   [blank, blank],
                   [color, color],
                   [blank, blank],
                   [color, color],
                   [blank, blank],
                   [color, color]]

    # Plot table
    ax.axis('tight')
    ax.axis('off')
    table = ax.table(text, loc='center', cellColours=cellColours, edges="closed")
    table.auto_set_font_size(False)
    table.set_fontsize(font_size)

    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()

    return ax


def mean_window(window_scores, center_x=None, title=None, xlabel=None, ylabel=None, show=True, save=None, ax=None):
    """
    """
    
    # Creat Axes object if not given
    if ax == None:
        with sns.axes_style("ticks"):
            fig, ax = plt.subplots(figsize=(7,5), tight_layout=True)
    
    # Plot metrics
    ax.plot(window_scores, color="black", linewidth=1.0)
    sns.despine(trim=True)

    # Plot center
    if center_x is not None:
        ax.axvline(x=center_x, linestyle="--", color="grey", linewidth=1.0, dash_capstyle="round")
        ax.set_xticks([0, center_x, len(window_scores)])
        ax.set_xticklabels([-center_x, 0, len(window_scores)-center_x])
    
    # Set labels
    if title is not None:
        ax.set_title(title, fontsize=14, fontweight='bold')
    if xlabel is not None:
        ax.set_xlabel(xlabel, fontsize=12)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=12)
    
    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()


def plot_genome_signal(df, signal_label, chrom_shift, line=True, plot_median=True, smooth=False, title=None,
                       ylabel=None, show=True, save=None, ax=None):
    """
    Plot a signal across the genome
    """

    # Create Axes object if not given
    if ax == None:
        with sns.axes_style("ticks"):
            fig, ax = plt.subplots(figsize=(10,5), tight_layout=True)

    # Iterate over chroms
    chrom_index = {chrom:df.loc[:,"seqnames"].values==chrom for chrom in np.unique(df.loc[:,"seqnames"].values)}
    for chrom in chrom_index:
        signal = df.loc[chrom_index[chrom], :]
        # Smooth signal
        if smooth:
            values = gaussian_smooth(signal.loc[:,signal_label].values, scale=10)
        else:
            values = signal.loc[:,signal_label].values
        shift = chrom_shift[chrom]
        if line:
            ax.plot(signal.loc[:,"start"].values+shift, values, linewidth=0.5, color="black")
        else:
            ax.scatter(signal.loc[:,"start"].values+shift, values, s=1, color="black")
        ax.axvline(x=shift, color="grey", linestyle="--")

    # Plot 0 line
    if plot_median:
        ax.axhline(y=np.median(df.loc[:,signal_label]), color="darkgrey", linewidth=0.5)
    
    # Set titles
    if title != None:
        ax.set_title(title, fontsize=14, fontweight='bold')
    ax.set_xlabel("Chromosomes", fontsize=12)
    if ylabel is not None:
        ax.set_ylabel(ylabel, fontsize=12)
    
    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()


def get_chrom_shift(chrom_lengths, chrom_order=None):
    """
    """

    chrom_shift = {}
    shift = 0

    if chrom_order is None:
        chrom_order = list(chrom_lengths.keys())

    for chrom in chrom_order:
        chrom_shift[chrom] = shift
        shift += chrom_lengths[chrom]

    return chrom_shift


def plot_cnv(pf: ProjectFrame,
             obs: str,
             cnv_column: str = "median",
             title: str | None = None,
             show: bool = True,
             save: str | None = None,
             plot_max: int = 1.2,
             plot_min: int = -1.2,
             ax = None):
    """
    Plot the CNVs
    """

    # Create Axes object if not given
    if ax == None:
        with sns.axes_style("ticks"):
            fig, ax = plt.subplots(figsize=(10,5), tight_layout=True)

    # Assign variables
    cnvs = pf.obs_intervals[obs]["cnv_segments"]
    bins = pf.intervals["cnv_bins"]

    # Determine whether chroms are
    chroms = cnvs.index.unique_labels

    # Get chrom shift
    chrom_shift = get_chrom_shift(pf.uns["chrom_lengths"], chroms)

    # Determine last position
    last_chrom = chroms[-1]
    last_position = chrom_shift[last_chrom]
    last_position += int(bins.loc[last_chrom,:].index.ends[-1]) + 10

    # Iterate over chroms
    for chrom in chroms:
        chrom_bins = bins.loc[chrom, obs]
        shift = chrom_shift[chrom]
        ax.scatter(chrom_bins.index.starts + shift, chrom_bins.values, s=1, color="black")
        ax.axvline(x=shift, color="grey", linestyle="--", linewidth=1.0, solid_capstyle="round")

    # Plot segments
    cnv_types = {"NEUT":"grey", "GAIN":"red", "HETD":"blue", "AMP":"darkred"}
    for i in range(cnvs.shape[0]):
        segment = cnvs.iloc[i,:]
        cnv_type = segment.df.loc[:,"Corrected_Call"].values[0]
        color =  cnv_types[cnv_type] if cnv_type in cnv_types else "darkred"
        start = segment.index.starts[0] + chrom_shift[segment.index.labels[0]]
        end = segment.index.ends[0] + chrom_shift[segment.index.labels[0]]
        ax.plot([start, end],
                  [segment.loc[:,"median"].values, segment.loc[:,"median"].values],
                  color=color, solid_capstyle="butt", linewidth=2.5)

    # Plot 0 line
    ax.axhline(y=0, color="darkgrey", linewidth=1.5)

    # Plot chromosome names
    text_x = []
    for chrom in chroms:
        text_x.append(chrom_shift[chrom] + (pf.uns["chrom_lengths"][chrom] / 2))
    ax.set_xticks(text_x)
    ax.set_xticklabels(chroms, rotation=90)
    
    # Set titles
    if title != None:
        ax.set_title(title, fontsize=14, fontweight='bold')
    ax.set_xlabel("Chromosomes", fontsize=12)
    ax.set_ylabel("log2 ratio", fontsize=12)
    ax.set_ylim((plot_min, plot_max))
    ax.set_xlim((0,last_position))

    # Plot arrows if segments of the chart
    off_chart = np.logical_or(cnvs.df.loc[:,"median"].values >= plot_max,
                              cnvs.df.loc[:,"median"].values <= plot_min)
    if np.sum(off_chart) > 0:
        off_segments = cnvs.loc[off_chart,:]
        for i in range(off_segments.shape[0]):
            segment = off_segments.iloc[i,:]
            cnv_type = segment.df.loc[:,"Corrected_Call"].values[0]
            color =  cnv_types[cnv_type] if cnv_type in cnv_types else "darkred"
            start = segment.starts[0] + chrom_shift[segment.index.labels[0]]
            start = start / last_position # convert to ratio
            end = segment.index.ends[0] + chrom_shift[segment.index.labels[0]]
            end = end / last_position # convert to ratio
            position = (start + end) / 2.0 # take middle
            median = segment.loc[:,"median"]

            if median >= 5.0:
                ax.annotate('', xy=(position, 1.01), xycoords='axes fraction', xytext=(position, 1.075),
                            arrowprops=dict(arrowstyle="simple", color=color))

            else:
                ax.annotate('', xy=(position, -0.01), xycoords='axes fraction', xytext=(position, -0.075),
                            arrowprops=dict(arrowstyle="simple", color=color))
    
    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()

    return ax


def fragment_distribution(pf, obs, title=None, show=True, save=None, ax=None):
    """
    Plot the fragment distribution
    """

    # Assign length dist
    length_dist = pf.obs_values["length_dist"].loc[obs,:].values
    
    # Determine total distribution
    if isinstance(length_dist, dict):
        max_length = max(len(length_dist[chrom]) for chrom in length_dist)
        fragment_lengths = np.zeros(max_length, dtype=int)
        for chrom in length_dist:
            fragment_lengths[:len(length_dist[chrom])] = length_dist[chrom]
    else:
        fragment_lengths = length_dist

    # Create Axes object if not given
    if ax == None:
        with sns.axes_style("ticks"):
            fig, ax = plt.subplots(figsize=(7,5), tight_layout=True)
    
    # Plot fragment lengths
    plt.plot(fragment_lengths, color="black", linewidth=0.5)
    sns.despine(trim=True, ax=ax)

    # Plot local maxima
    for maxima in argrelmax(fragment_lengths, order=35)[0]:
        plt.axvline(x=maxima, color="grey", linestyle="--", dash_capstyle="round", linewidth=0.5)
    
    # Set titles
    if title != None:
        ax.set_title(title, fontsize=14, fontweight='bold')
    ax.set_xlabel("Read length", fontsize=12)
    ax.set_ylabel("Density", fontsize=12)
    
    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()


def summary(pf, key, show=True, save=None):
    """
    Plot summary metrics

    Params
    ------
        metrics
            dict
        show
            bool
        save
            bool
    
    Returns
    -------
        fig
            matplotlib.pyplot.figure
    """

    # Initialize matplotlib grid
    fig = plt.figure(figsize=(10, 7), constrained_layout=True)
    gs = fig.add_gridspec(3, 3)

    # Plot table
    table_ax = fig.add_subplot(gs[0, 0])
    #sam_fn = os.path.split(metrics["sam"])[-1]
    sam_fn = key
    #cnv_hmm_model = np.argmax(metrics.hmm_loglik.loc[:,"loglik"].values)
    purity = pf.anno.loc[key,"purity"]
    ploidy = pf.anno.loc[key,"ploidy"]
    clonal = pf.anno.loc[key,"clonal"]
    n_frags = pf.anno.loc[key,"n_fragments"]
    nfr_score = pf.anno.loc[key,"long_hg19_TSS_nfr_score"]
    median_var = np.median(pf.obs_intervals["cnv_segments"][key].loc[:,"var"].values)

    table_ax = plot_table(sam_fn, pf.ref_genome, pf.cnv_binsize,
                          purity, ploidy, clonal, n_frags, nfr_score, median_var,
                          font_size = 5, show=False, ax=table_ax)

    # Plot fragment length distribution
    with sns.axes_style("ticks"):
        len_ax = fig.add_subplot(gs[0, 1])
    len_ax = fragment_distribution(pf, key, title=None, show=False, ax=len_ax)
    
    # Plot TSS profile
    with sns.axes_style("ticks"):
        tss_ax = fig.add_subplot(gs[0, 2])
    mean_wps = pf.obs_values["long_hg19_TSS_mean_wps"].loc[key,:].values
    tss_ax = mean_window(mean_wps, center_x=1000, title=None,
                         ylabel="Normalized WPS", xlabel="Distance to TSS",
                         show=False, ax=tss_ax)

    # Plot CNV plot
    with sns.axes_style("ticks"):
        cnv_ax = fig.add_subplot(gs[1:3,:])
    cnv_ax = plot_cnv(pf, key, title=None, show=False, ax=cnv_ax)

    # Plot WPS peak distances
    #with sns.axes_style("ticks"):
        #peak_dist_ax = fig.add_subplot(gs[2,:])
    #peak_dist_ax = plot_genome_signal(metrics["peak_dist"], "peak_dist", metrics.chrom_shift, line=True, smooth=True, ax=peak_dist_ax, ylabel="WPS peak distances", show=False)

    # Save of display plot
    if save != None:
        plt.savefig(save, transparent=True, dpi=80)
    if show:
        plt.show()

    return fig


from intervalframe import IntervalFrame
import ngsfragments as ngs


def sequencing_call_cnvs(data: IntervalFrame,
                         genome_version: str = "hg38",
                         n_probes: int = None,
                         n_probes_hmm: int = None,
                         zscore: bool = False,
                         n_mads: float = 1.48) -> ngs.segment.CNVcaller:
    """
    """

    if n_probes is None:
        n_probes = data.shape[0] // 40000
        n_probes_hmm = data.shape[0] // 20000

        # Minimum number of probes
        if n_probes < 10:
            n_probes = 10
        if n_probes_hmm < 15:
            n_probes_hmm = 15

    # Call CNVs
    cnvs = ngs.segment.CNVcaller(genome_version = genome_version,
                                        scStates = None,
                                        n_per_bin = n_probes,
                                        n_per_bin_hmm = n_probes_hmm,
                                        n_mads = n_mads)
    
    cnvs.predict_cnvs(data = data)

    # Calculate zscore
    if zscore:
        cnvs.calculate_zscore()

    return cnvs


from projectframe import ProjectFrame
import ngsfragments as ngs
import glob
import pandas as pd
import numpy as np
import os

# Local imports
from .seq_utilities import read_methyldackel_bed
from .read_methyldackel import read_methyldackel                       


def read_betas(meth_files: str,
               use_sparse: bool = False):
    """
    Read methyldackel files

    Parameters
    ----------
        meth_files : str
            Methyldackel files
        use_sparse : bool = False
            Whether to use sparse optimization

    Returns
    -------
        all_cpgs : IntervalFrame
            IntervalFrame containing all CpGs
    """

    # Find sample name
    meth_names = np.array([os.path.split(file)[-1].split(".bedGraph")[0] for file in meth_files])
    # Remove CpG string
    meth_names = np.array([name.split("_CpG")[0] for name in meth_names])

    # Iterate through files
    for i, file in enumerate(meth_files):
        meth_iframe = read_methyldackel_bed(file)
        if i == 0:
            all_cpgs = meth_iframe
        else:
            all_cpgs = all_cpgs.combine([meth_iframe.index], sparse_optimize=use_sparse)
    
    return all_cpgs


def process_sequencing(methyldackel_directory: str,
                       bam_directory: str = None,
                       genome_version: str = "hg19",
                       nthreads: int = 1,
                       call_cnvs: bool = False,
                       use_sparse: bool = False,
                       verbose: bool = False):
    """
    Process sequencing data

    Parameters
    ----------
        methyldackel_directory : str
            Directory containing methyldackel files
        bam_directory : str = None
            Directory containing bam files
        genome_version : str = "hg19"
            Which genome version to use
        nthreads : int = 1
            Number of threads to use
        call_cnvs : bool = False
            Whether to call copy number variations
        verbose : bool = False
            Whether to print progress

    Returns
    -------
        pf : ProjectFrame
            Object containing results
    """

    # Import genome
    if genome_version == "hg19":
        from hg19genome import Hg19Genome
        Genome = Hg19Genome()
    elif genome_version == "hg38":
        from hg38genome import Hg38Genome
        Genome = Hg38Genome()
    else:
        raise NotImplementedError("Other genome versions not implemented...yet.")

    # Initialize projectframe
    pf = ProjectFrame()
    pf.params["ref_genome"] = genome_version

    # Find files
    meth_files = glob.glob(os.path.join(methyldackel_directory, "*.bedGraph"))
    # Find sample name
    meth_names = np.array([os.path.split(file)[-1].split(".bedGraph")[0] for file in meth_files])
    # Remove CpG string
    meth_names = np.array([name.split("_CpG")[0] for name in meth_names])

    # Read methyldackel
    if verbose: print("Reading methyldackel files", flush=True)
    betas = read_methyldackel(np.array(meth_files).astype(str), use_sparse=use_sparse)
    pf.add_intervals("betas", betas)
    pf.add_anno("methyl_type", value="Sequencing")

    # Read bams
    if call_cnvs and bam_directory is not None:
        bam_files = glob.glob(os.path.join(bam_directory, "*.bam"))
        # Find sample name
        bam_names = np.array([os.path.split(file)[-1].split(".bam")[0] for file in bam_files])
        if verbose: print("Reading bams.")
        for i, bam in enumerate(bam_files):
            if verbose: print(bam)
            pf = ngs.segment.call_cnv_pipeline(pf, bam, genome_version=genome_version, verbose=verbose, nthreads=nthreads, wgbs=True)
            del pf.uns.dict[bam_names[i]]
            pf.obs_intervals[bam_names[i]]["cnv_segments"].drop_columns(["copy_number","event","subclone_status","logR_Copy_Number"])

    return pf


import numpy as np
import pandas as pd
from scipy import stats
from hmmlearn import hmm
from ailist import LabeledIntervalArray
from intervalframe import IntervalFrame


def detect_fhrs_hmm_based(beta_values, positions, chromosome,
                          min_sites=3, max_length=1000, n_states=3,
                          pvalue_threshold=0.05):
    """
    Detect focally hypomethylated regions using HMM for initial segmentation
    and then applying additional filtering criteria.
    
    Parameters:
    -----------
    beta_values : array-like
        Array of methylation beta values (0-1)
    positions : LabeledIntervalArray
        Genomic positions corresponding to beta values
    min_sites : int
        Minimum number of sites required in a hypomethylated region
    max_length : int
        Maximum length of a hypomethylated region in base pairs
    n_states : int
        Number of methylation states for HMM
    pvalue_threshold : float
        Threshold for statistical significance when comparing to surrounding regions
        
    Returns:
    --------
    list
        List of tuples (start_idx, end_idx, pvalue, avg_methylation, length) for each FHR
    """
    
    starts = positions.starts
    ends = positions.ends
    
    # 1. Use HMM to segment the methylation data
    X = np.array(beta_values).reshape(-1, 1)
    model = hmm.GaussianHMM(n_components=n_states, covariance_type="full", 
                          n_iter=100, random_state=42)
    model.fit(X)
    state_sequence = model.predict(X)
    
    # 2. Determine which state represents hypomethylated regions
    state_means = np.array([model.means_[i][0] for i in range(n_states)])
    hypomethylated_state = np.argmin(state_means)
    
    # 3. Find contiguous regions of the hypomethylated state
    candidates = []
    start_idx = None
    
    for i, state in enumerate(state_sequence):
        if state == hypomethylated_state and start_idx is None:
            start_idx = i
        elif state != hypomethylated_state and start_idx is not None:
            candidates.append((start_idx, i-1))
            start_idx = None
    
    # Handle case where the last region is hypomethylated
    if start_idx is not None:
        candidates.append((start_idx, len(state_sequence)-1))
    
    # 4. Filter candidates based on criteria
    fhrs = []
    fnhrs_intervals = LabeledIntervalArray()
    
    for start_idx, end_idx in candidates:
        # Check region length constraint
        if positions is not None:
            region_length = ends[end_idx] - starts[start_idx]
            if region_length > max_length:
                continue
        
        # Check minimum number of sites
        site_count = end_idx - start_idx + 1
        if site_count < min_sites:
            continue
        
        # Statistical test: compare with surrounding regions
        # Define surrounding regions (equal total length to focal region)
        region_size = end_idx - start_idx + 1
        
        # Left surrounding region
        left_surround_start = max(0, start_idx - region_size)
        left_surround_end = start_idx
        left_surround = beta_values[left_surround_start:left_surround_end]
        
        # Right surrounding region
        right_surround_start = end_idx + 1
        right_surround_end = min(len(beta_values), end_idx + 1 + region_size)
        right_surround = beta_values[right_surround_start:right_surround_end]
        
        # Combine surrounding regions
        surrounding = np.concatenate([left_surround, right_surround])
        
        # Focal region
        focal = beta_values[start_idx:end_idx+1]
        
        # Statistical test (Mann-Whitney U test)
        _, pvalue = stats.mannwhitneyu(surrounding, focal, alternative='greater')
        
        if pvalue < pvalue_threshold:
            # Calculate average methylation in the region
            avg_methylation = np.mean(focal)
            
            fhrs.append((pvalue, avg_methylation, site_count))
            fnhrs_intervals.add(starts[start_idx], ends[end_idx] + 1, chromosome)

    # Convert to IntervalFrame
    if len(fhrs) == 0:
        return IntervalFrame()
    
    fhrs = pd.DataFrame(np.array(fhrs))
    #print(fhrs, flush=True)
    fhrs = IntervalFrame(df=fhrs, intervals=fnhrs_intervals)
    fhrs.df.columns = ["pvalue", "avg_methylation", "length"]
    
    return fhrs


def detect_lmrs(betas: IntervalFrame,
                sample: str,
                min_sites: int = 3,
                max_length: int = 1000,
                n_states: int = 3,
                pvalue_threshold: float = 0.05) -> IntervalFrame:
    """
    Detect focally hypermethylated regions (LMRs) using HMM for initial segmentation
    and then applying additional filtering criteria.
    
    Parameters:
    -----------
    betas : IntervalFrame
        Methylation data
    sample : str
        Sample name
    min_sites : int
        Minimum number of sites required in a hypermethylated region
    max_length : int
        Maximum length of a hypermethylated region in base pairs
    n_states : int
        Number of methylation states for HMM
    pvalue_threshold : float
        Threshold for statistical significance when comparing to surrounding regions
        
    Returns:
    --------
    IntervalFrame
        List of tuples (start_idx, end_idx, pvalue, avg_methylation, length) for each LMR
    """
    
    # Extract beta values and positions from the IntervalFrame
    lmr_results = []
    chromosomes = betas.index.unique_labels
    for chrom in chromosomes:
        print("Processing chromosome:", chrom, flush=True)
        
        # Extract beta values and positions
        beta_values = betas.loc[chrom,:].df.loc[:,sample].values
        positions = betas.loc[chrom,:].index
        
        if len(beta_values) < 100:
            print("Skipping chromosome", chrom, "due to insufficient data", flush=True)
            continue
        # Call the detect_fhrs_hmm_based function to find LMRs
        lmrs = detect_fhrs_hmm_based(beta_values, positions, chrom,
                                                 min_sites, max_length, n_states,
                                                 pvalue_threshold)
        if lmrs.shape[0] > 0:
            lmr_results.append(lmrs)
    
    # Call the detect_fhrs_hmm_based function to find LMRs
    lmr_results = IntervalFrame.combine(lmr_results)
    lmr_results.df.columns = ["pvalue", "avg_methylation", "length"]
    
    return lmr_results


import numpy as np
import pandas as pd
from projectframe import ProjectFrame
from intervalframe import IntervalFrame

# Local imports
from ...core.utilities import bin_cpgs


def tile_methylation(pf: ProjectFrame,
                     bin_size: int = 100000,
                     genome_version: str = "hg19",
                     remove_nans: bool = True,
                     method: str = "mean") -> ProjectFrame:
    """
    Tile methylation values into bins.

    Parameters
    ----------
        pf : ProjectFrame
            ProjectFrame of CpG methylation values.
        bin_size : int
            Number of bins to tile into.
        genome_version : str
            Genome version to use.
        remove_nans : bool
            If True, remove bins with NaNs.
        method : str
            Method to use for tiling.

    Returns
    -------
        pf : ProjectFrame
    
    """

    # Assign genome
    import genome_info
    genome = genome_info.GenomeInfo(genome_version)

    # Get bins
    bins = IntervalFrame.from_dict_range(genome["chrom_sizes"], bin_size = bin_size)

    # Tile betas
    bins = bin_cpgs(pf.intervals["betas"], bins=bins, method=method)

     # Remove NaNs
    if remove_nans:
        bins = bins.iloc[np.sum(pd.isnull(bins.values), axis=1) == 0,:]

    # Assign to projectframe
    pf.add_intervals("tile_methylation", bins)

    return pf


def region_methylation(pf: ProjectFrame,
                        feature: str = "gene",
                        upstream: int = 1000,
                        downstream: int = 1000,
                        gene_type: str = "protein_coding",
                        genome_version: str = "hg19",
                        remove_nans: bool = True,
                        method: str = "mean") -> ProjectFrame:
    """
    Region methylation values into bins.

    Parameters
    ----------
        pf : ProjectFrame
            ProjectFrame of CpG methylation values.
        bin_size : int
            Number of bins to tile into.
        genome_version : str
            Genome version to use.
        remove_nans : bool
            Remove NaNs from the resulting ProjectFrame.
        method : str
            Method to use for tiling.

    Returns
    -------
        pf : ProjectFrame
    
    """

    # Assign genome
    import genome_info
    genome = genome_info.GenomeInfo(genome_version)

    # Get regions
    if feature == "gene":
        genes = genome.get_intervals("gene", upstream = 5000, filter_column = "gene_type", filter_selection="protein_coding")
        names = genes.df.loc[:,"gene_name"].values
    elif feature == "promoter":
        genes = genome.get_intervals("tss", upstream = 5000, downstream=1000, filter_column = "gene_type", filter_selection="protein_coding")
        names = genes.df.loc[:,"gene_name"].values
    else:
        raise NotImplementedError("Only gene and promoter are supported currently")

    # Remove unnecessary columns
    genes.drop_columns(["Strand","gene_type","level","Source"])

    # Tile betas
    bins = bin_cpgs(pf.intervals["betas"], bins=genes, method=method)

     # Remove NaNs
    if remove_nans:
        selected = np.sum(pd.isnull(bins.values), axis=1) == 0
        names = names[selected]
        bins = bins.iloc[selected,:]
    bins.df.index = names
    bins = bins.df.T

    # Append to projectframe
    pf.add_values(feature+"_methylation", bins)

    return pf


def binarize_methylation(pf: ProjectFrame,
                         key: str = "betas",
                         new_key: str = "binary_methylation",
                         threshold: float = 0.6,
                         use_sparse: bool = False) -> ProjectFrame:
    """
    """

    # Copy betas
    betas = pf.intervals["betas"].copy()
    
    # Binarize
    betas.df[betas.df < threshold] = -1
    betas.df[betas.df >= threshold] = 1
    betas.df[pd.isnull(betas.df)] = 0
    if use_sparse:
        betas.df = betas.df.astype(pd.SparseDtype(int, 0))
    else:
        betas.df = betas.df.astype(int)

    # Assign to projectframe
    pf.add_intervals("binary_betas", betas)

    return pf






import numpy as np
import pandas as pd
import json
from scipy.special import softmax
from ..Architectures.MultiClassEnsemble import EnsembleModel
import torch


def logsoftmax(x):
    c = np.max(x, axis=1, keepdims=True)  # Max per row for numerical stability
    logsumexp = np.log(np.sum(np.exp(x - c), axis=1, keepdims=True))  # Log-sum-exp per row
    return x - c - logsumexp  # Log-softmax formula


class MPACT_classifier(object):
    """
    """
    
    def __init__(self,
                 directory,
                 threshold = 0.7):
        """
        """

        self.device = torch.device("cpu")
        self.models = [EnsembleModel.load(directory+"MPACT_D1_v1.pth", device=self.device),
                       EnsembleModel.load(directory+"MPACT_S1_v1.pth", device=self.device),
                       EnsembleModel.load(directory+"MPACT_S2_v1.pth", device=self.device)]
        self.features1 = np.loadtxt(directory+"Features1_v4.txt", dtype=str)
        self.features2 = np.loadtxt(directory+"Features2_v4.txt", dtype=str)
        self.features3 = np.loadtxt(directory+"Features3_v4.txt", dtype=str)

        self.codings = json.load(open(directory+"codings.json"))
        self.class_merge = self.codings["merge"]
        self.decoder = self.codings["decoder"]

        self.threshold = threshold
    

    def predict_model(self,
                      input1,
                      input2,
                      input3,
                      model_number,
                      merge = True):
        """
        """
        # Load model
        model = self.models[model_number - 1]

        # Check input shapes
        dim = 2
        if input1.shape[0] == 1:
            dim = 1
            # Add dummy rows for single sample input
            input1 = pd.DataFrame(np.repeat(input1, 2, axis=0), 
                                  index=[input1.index.values[0], input1.index.values[0]],
                                  columns=input1.columns)
            input2 = pd.DataFrame(np.repeat(input2, 2, axis=0), 
                                  index=[input2.index.values[0], input2.index.values[0]],
                                  columns=input2.columns)
            input3 = pd.DataFrame(np.repeat(input3, 2, axis=0), 
                                  index=[input3.index.values[0], input3.index.values[0]],
                                  columns=input3.columns)

        probabilities, predictions = model.predict_step({"x": torch.from_numpy(input1.values.astype(np.float32))},
                                                        {"x": torch.from_numpy(input2.values.astype(np.float32))},
                                                        {"x": torch.from_numpy(input3.values.astype(np.float32))})
        probabilities = probabilities.detach().cpu().numpy()

        # Reshape probabilities if single sample input
        if dim == 1:
            probabilities = probabilities[[0],:]
            predictions = predictions[[0]]
            input1 = input1.iloc[[0],:]
            input2 = input2.iloc[[0],:]
            input3 = input3.iloc[[0],:]

        # Get classes
        classes = pd.Series(self.decoder).values

        # Merge classes
        probabilities = pd.DataFrame(probabilities, index=input1.index.values, columns=classes)
        if merge:
            for c in self.class_merge:
                probabilities.loc[:,c] = probabilities.loc[:,self.class_merge[c]].sum(axis=1).values
                probabilities = probabilities.drop(columns=list(set(self.class_merge[c]) - set([c])))
        
        # Determine predictions
        predictions = probabilities.columns.values[probabilities.values.argmax(axis=1)]
        
        return predictions, probabilities
        
        
    def predict(self,
                X: pd.DataFrame,
                method: str = "max",
                transform: bool = True):
        """
        """

        # Scale X
        if transform:
            X = (X * 2) - 1  # Scale to [-1, 1] for MLP input

        # Divide X into feature folds
        input1 = X.loc[:,self.features1]
        input2 = X.loc[:,self.features2]
        input3 = X.loc[:,self.features3]
        
        # Get dense predictions
        dense_predictions, dense_probabilities = self.predict_model(input1, input2, input3, 1, merge=True)
        
        # Get sparse model 1 predictions
        sparse1_predictions, sparse1_probabilities = self.predict_model(input1, input2, input3, 2, merge=True)
        
        # Get sparse model 2 predictions
        sparse2_predictions, sparse2_probabilities = self.predict_model(input1, input2, input3, 3, merge=True)

        # Determine probabilities
        if method == "average":
            probabilities = (dense_probabilities + sparse1_probabilities + sparse2_probabilities) / 3
        elif method == "max":
            probabilities = dense_probabilities.copy()
            # Iterate over sparse models and take max probabilities
            for i in range(sparse1_probabilities.shape[0]):
                if sparse1_probabilities.values[i,:].max() > probabilities.values[i,:].max():
                    probabilities.iloc[i,:] = sparse1_probabilities.values[i,:]
                if sparse2_probabilities.values[i,:].max() > probabilities.values[i,:].max():
                    probabilities.iloc[i,:] = sparse2_probabilities.values[i,:]
        elif method == "concat":
            probabilities = pd.concat([dense_probabilities, sparse1_probabilities, sparse2_probabilities]).groupby(level=0).max()
            # Sum to 1
            probabilities = probabilities.div(probabilities.sum(axis=1), axis=0)
        else:
            raise ValueError("Method not recognized.")
        
        # Determine predictions
        probabilities = probabilities.loc[X.index.values,:]
        predictions = probabilities.columns.values[probabilities.values.argmax(axis=1)]

        return predictions, probabilities